/**
 * Copyright (C) 2006 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.inject;

import com.google.inject.InjectorImpl.SingleMemberInjector;
import com.google.inject.InjectorImpl.SingleParameterInjector;
import com.google.inject.internal.ErrorMessage;
import com.google.inject.internal.ResolveFailedException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;

/**
 * Injects constructors.
 *
 * @author crazybob@google.com (Bob Lee)
 */
class ConstructorInjector<T> {

  final Class<T> implementation;
  final SingleMemberInjector[] memberInjectors;
  final SingleParameterInjector<?>[] parameterInjectors;
  final ConstructionProxy<T> constructionProxy;

  ConstructorInjector(InjectorImpl injector, Class<T> implementation) {
    this.implementation = implementation;
    constructionProxy = injector.reflection.getConstructionProxy(implementation);
    parameterInjectors = createParameterInjector(injector, constructionProxy);
    List<SingleMemberInjector> memberInjectorsList = injector.injectors.get(implementation);
    memberInjectors = memberInjectorsList.toArray(
        new SingleMemberInjector[memberInjectorsList.size()]);
  }

  SingleParameterInjector<?>[] createParameterInjector(
      InjectorImpl injector, ConstructionProxy<T> constructionProxy) {
    try {
      return constructionProxy.getParameters().isEmpty()
          ? null // default constructor.
          : injector.getParametersInjectors(
              constructionProxy.getMember(),
              constructionProxy.getParameters());
    }
    catch (ResolveFailedException e) {
      injector.errorHandler.handle(e.getMessage(constructionProxy.getMember()));
      return null;
    }
  }

  /**
   * Construct an instance. Returns {@code Object} instead of {@code T} because
   * it may return a proxy.
   */
  Object construct(InternalContext context, Class<?> expectedType) {
    ConstructionContext<T> constructionContext
        = context.getConstructionContext(this);

    // We have a circular reference between constructors. Return a proxy.
    if (constructionContext.isConstructing()) {
      // TODO (crazybob): if we can't proxy this object, can we proxy the
      // other object?
      return constructionContext.createProxy(expectedType);
    }

    // If we're re-entering this factory while injecting fields or methods,
    // return the same instance. This prevents infinite loops.
    T t = constructionContext.getCurrentReference();
    if (t != null) {
      return t;
    }

    try {
      // First time through...
      constructionContext.startConstruction();
      try {
        Object[] parameters
            = InjectorImpl.getParameters(context, parameterInjectors);
        t = constructionProxy.newInstance(parameters);
        constructionContext.setProxyDelegates(t);
      }
      finally {
        constructionContext.finishConstruction();
      }

      // Store reference. If an injector re-enters this factory, they'll
      // get the same reference.
      constructionContext.setCurrentReference(t);

      // Inject fields and methods.
      for (InjectorImpl.SingleMemberInjector injector : memberInjectors) {
        injector.inject(context, t);
      }

      return t;
    }
    catch (InvocationTargetException e) {
      Throwable cause = e.getCause() != null ? e.getCause() : e;
      throw new ProvisionException(ErrorMessage.errorInjectingConstructor().toString(), cause);
    }
    finally {
      constructionContext.removeCurrentReference();
    }
  }
}
